package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> /*extends PrintableTree<Forest.Body<T>>*/ implements Iterable<Forest.Body<T>> {

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public final Tree<T> expand1() throws Ambiguous, ParseFailed {
        Iterator<Tree<T>> it = expand(true).iterator();
        if (!it.hasNext()) throw new ParseFailed();
        return it.next();
    }

    /** expand this forest into a set of trees */
    public HashSet<Tree<T>> expand(boolean toss) {
        HashSet<Tree<T>> ret = new HashSet<Tree<T>>();
        for(Body<T> b : this)
            ret.addAll(b.expand(toss, new ArrayList<Tree<T>>(), 0, new HashSet<Tree<T>>()));
        if (toss && ret.size() > 1) throw new Ambiguous(this);
        return ret;
    }

    static        <T> Forest<T> singleton(Input.Location loc)                       { return create(loc, null, new Forest[] { }, false, true); }
    static        <T> Forest<T> singleton(Input.Location loc, Forest<T> body)       { return create(loc, null, new Forest[] { body },  false, true); }
    static        <T> Forest<T> leaf(Input.Location loc, T tag) { return create(loc, tag, null, false, false); }
    public static <T> Forest<T> create(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton) {
        return new MyBody<T>(loc, tag, tokens, unwrap, singleton);
    }

    // Body //////////////////////////////////////////////////////////////////////////////

    protected static interface Body<T> {
         HashSet<Tree<T>> expand(boolean toss, ArrayList<Tree<T>> toks, int i, HashSet<Tree<T>> h);
    }

    protected static class MyBody<T> extends Forest<T> implements Body<T> /* extends PrintableTree<Forest<T>> implements Iterable<Forest.Body<T>>*/ {
        public Iterator<Body<T>> iterator() { return new SingletonIterator<Body<T>>(this); }

        private final Input.Location    location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final boolean           unwrap;
        private final boolean           singleton;

        private MyBody(Input.Location loc, T tag, Forest<T>[] tokens, boolean unwrap, boolean singleton) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            if (tokens != null) for(int i=0; i<tokens.length; i++) if (tokens[i]==null) throw new Error(i+"");
            this.unwrap = unwrap;
            this.singleton = singleton;
        }

        public HashSet<Tree<T>> expand(boolean toss, ArrayList<Tree<T>> toks, int i, HashSet<Tree<T>> h) {
            if (singleton) {
                for(Body<T> b : tokens[0]) b.expand(toss, toks, i, h);

            } else if (i==tokens.length) {
                h.add(new Tree<T>(null, tag, toks.toArray(tree_hint)));

            } else if (unwrap && i==tokens.length-1) {
                if (tokens[i] != null)
                    for(Body b : tokens[i])
                        b.expand(toss, toks, 0, h);

            } else {
                boolean hit = false;
                for(Tree<T> r : tokens[i].expand(toss)) {
                    hit = true;
                    int old = toks.size();
                    toks.add(r);
                    expand(toss, toks, i+1, h);
                    while(toks.size() > old) toks.remove(toks.size()-1);
                }
                //if (!hit) throw new Error();
            }
            return h;
        }

        protected String  headToString()         { return null; }
        protected String  headToJava()           { return null; }
        protected String  left()                 { return "{"; }
        protected String  right()                { return "}"; }
        protected boolean ignoreSingleton()      { return false; }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends Forest<T> {
        private FastSet<Forest<T>> hp = new FastSet<Forest<T>>();
        public Ref() { }
        public void merge(Forest p) { if (p!=this) hp.add(p, true); }
        public Iterator<Body<T>> iterator() {
            final Iterator<Forest<T>> ift = hp==null ? null : hp.iterator();
            return new Iterator<Body<T>>() {
                Iterator<Body<T>> ibt = ift==null ? null : ift.hasNext() ? ift.next().iterator() : null;
                public void remove() { throw new RuntimeException("not supported"); }
                public boolean hasNext() {
                    if (ibt==null) return false;
                    if (ibt.hasNext()) return true;
                    ibt = ift.hasNext() ? ift.next().iterator() : null;
                    return hasNext();
                }
                public Body<T> next() {
                    return ibt.next();
                }
            };

        }
        public Forest resolve() { return this; }
    }


    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static final Forest[] emptyForestArray = new Forest[0];

    protected String  headToString()    { return null; }
    protected String  headToJava()      { return null; }
    protected String  left()            { return "<?"; }
    protected String  right()           { return "?>"; }
    protected boolean ignoreSingleton() { return true; }
}
